/*
** maplabel.c: Routines to enable text drawing, BITMAP or TRUETYPE.
*/
#include <gdfonts.h>
#include <gdfontl.h>
#include <gdfontt.h>
#include <gdfontmb.h>
#include <gdfontg.h>

#ifdef USE_TTF
#include <gdcache.h>
#include <gdttf.h>
#include "freetype.h"
#endif

#include "map.h"

int msAddLabel(mapObj *map, int layer, int class, int tile, int shape, pointObj labelPoint, char *string, int size, double angle, double feature_size)
{
  int i;

  if(!string)
    return(0); /* not an error */

  if(map->labelcache.numlabels == map->labelcache.cachesize) { /* just add it to the end */
    map->labelcache.labels = (labelCacheMemberObj *) realloc(map->labelcache.labels, sizeof(labelCacheMemberObj)*(map->labelcache.cachesize+MS_LABELCACHEINCREMENT));    
    if(map->labelcache.labels == NULL) {
      msSetError(MS_MEMERR, "Realloc() error.", "msAddLabel()");
      return(-1);
    }   
    map->labelcache.cachesize+=MS_LABELCACHEINCREMENT;
  }

  i = map->labelcache.numlabels;  
  map->labelcache.labels[i].layeridx = layer;
  map->labelcache.labels[i].classidx = class;
  map->labelcache.labels[i].tileidx = tile;
  map->labelcache.labels[i].shapeidx = shape;
  map->labelcache.labels[i].point = labelPoint;
  map->labelcache.labels[i].string = strdup(string);
  map->labelcache.labels[i].size =  size;
  map->labelcache.labels[i].angle =  angle;
  map->labelcache.labels[i].featuresize = feature_size;

  map->labelcache.labels[i].poly = (shapeObj *) malloc(sizeof(shapeObj));
  map->labelcache.labels[i].poly->numlines = 0;
  map->labelcache.labels[i].poly->line = NULL;

  map->labelcache.labels[i].status = MS_FALSE;

  if(map->layers[layer].type == MS_POINT) { 
    rectObj rect;
    int w, h;

    msGetMarkerSize(&(map->markerset), &(map->layers[layer].class[class]), &w, &h);
    rect.minx = labelPoint.x - MS_NINT(.5 * w);
    rect.miny = labelPoint.y - MS_NINT(.5 * h);
    rect.maxx = labelPoint.x + MS_NINT(.5 * w);
    rect.maxy = labelPoint.y + MS_NINT(.5 * h);
    msRect2Polygon(rect, map->labelcache.labels[i].poly);

    map->labelcache.labels[i].status = MS_TRUE;
  }

  map->labelcache.numlabels++;
  
  return(0);
}

#ifdef USE_TTF
int msLoadFontSet(fontSetObj *fontset) 
{
  FILE *stream;
  char buffer[MS_BUFFER_LENGTH];
  char alias[64], file1[MS_PATH_LENGTH], file2[MS_PATH_LENGTH];
  char *path;
  int i;

  if(fontset->numfonts != 0) /* already initialized */
    return(0);

  if(!fontset->filename)
    return(0);

  path = getPath(fontset->filename);

  fontset->fonts = msCreateHashTable(); /* create font hash */  
  if(!fontset->fonts) {
    msSetError(MS_HASHERR, "Error initializing font hash.", "msLoadFontset()");
    return(-1);
  }

  stream = fopen(fontset->filename, "r");
  if(!stream) {
    msSetError(MS_IOERR, NULL, "msLoadFontset()");
    sprintf(ms_error.message, "Error opening fontset %s.", fontset->filename);
    return(-1);
  }
 
  i = 0;
  while(fgets(buffer, MS_BUFFER_LENGTH, stream)) { /* while there's something to load */

    if(buffer[0] == '#' || buffer[0] == '\n' || buffer[0] == '\r' || buffer[0] == ' ')
      continue; /* skip comments and blank lines */

    sscanf(buffer,"%s %s", alias,  file1);

    if(file1[0] == '/') { /* already full path */      
      msInsertHashTable(fontset->fonts, alias, file1);
    } else {
      sprintf(file2, "%s%s", path, file1);      
      msInsertHashTable(fontset->fonts, alias, file2);
    }

    i++;
  }

  fontset->numfonts = i;
  fclose(stream); /* close the file */
  free(path);

  return(0);
}
#endif

/*
** Note: All these routines assume a reference point at the LL corner of the text. GD's
** bitmapped fonts use UL and this is compensated for.
*/

int msGetLabelSize(char *string, labelObj *label, int *w, int *h, fontSetObj *fontset) /* assumes an angle of 0 */
{

  if(label->type == MS_TRUETYPE) {
#ifdef USE_TTF
    int bbox[8];
    char *error=NULL, *font=NULL;

    font = msLookupHashTable(fontset->fonts, label->font);
    if(!font) {
      msSetError(MS_TTFERR, NULL, "msGetLabelSize()");
      sprintf(ms_error.message, "Requested font (%s) not found.", label->font);
      return(-1);
    }

    error = imageStringTTF(NULL, bbox, 0, font, label->size, 0, 0, 0, string, label->wrap);
    if(error) {
      msSetError(MS_TTFERR, error, "msGetLabelSize()");
      return(-1);
    }    

    /* compute w & h */
    *w = bbox[2] - bbox[0];
    *h = -bbox[5];
#else
    msSetError(MS_TTFERR, "TrueType font support is not available.", "msGetLabelSize()");
    return(-1);
#endif
  } else { /* MS_BITMAP font */
    gdFontPtr fontPtr;
    char **token=NULL;
    int t, num_tokens, max_token_length=0;

    if((fontPtr = msGetBitmapFont(label->size)) == NULL)
      return(-1);

    if((token = split(string, label->wrap, &(num_tokens))) == NULL)
      return(0);

    for(t=0; t<num_tokens; t++) /* what's the longest token */
      max_token_length = MS_MAX(max_token_length, strlen(token[t]));

    /* compute w & h */
    *w = fontPtr->w * max_token_length;
    *h = fontPtr->h * num_tokens;

    msFreeCharArray(token, num_tokens);
  }
  return(0);
}

gdFontPtr msGetBitmapFont(int size)
{
  switch(size) { /* set the font to use */
  case MS_TINY:
    return(gdFontTiny);
    break;
  case MS_SMALL:  
    return(gdFontSmall);
    break; 
  case MS_MEDIUM:
    return(gdFontMediumBold);
    break; 
  case MS_LARGE:
    return(gdFontLarge);
    break;  
  case MS_GIANT:
    return(gdFontGiant);
    break;
  default:
    msSetError(MS_GDERR, "Invalid bitmap font.", "msDrawLabel()");
    return(NULL);
  }
}

#define MARKER_SLOP 2

static pointObj get_metrics(pointObj *p, int position, int w, int h, int ox, int oy, double angle, int buffer, shapeObj *poly)
{
  pointObj q;
  double x1=0, y1=0, x2=0, y2=0;
  lineObj line={0,NULL};
  double sin_a,cos_a;

  switch(position) {
  case MS_UL:
    x1 = -w - ox;
    y1 = -oy;
    break;
  case MS_UC:
    x1 = -(w/2.0);
    y1 = -oy - MARKER_SLOP;
    break;
  case MS_UR:
    x1 = ox;
    y1 = -oy;
    break;
  case MS_CL:
    x1 = -w - ox - MARKER_SLOP;
    y1 = (h/2.0);
    break;
  case MS_CC:
    x1 = -(w/2.0);
    y1 = (h/2.0);
    break;
  case MS_CR:
    x1 = ox + MARKER_SLOP;
    y1 = (h/2.0);
    break;
  case MS_LL:
    x1 = -w - ox;
    y1 = h + oy;
    break;
  case MS_LC:
    x1 = -(w/2.0);
    y1 = h + oy + MARKER_SLOP;
    break;
  case MS_LR:
    x1 = ox;
    y1 = h + oy;
    break;
  }

  sin_a = sin(angle);
  cos_a = cos(angle);

  q.x = p->x + MS_NINT(x1 * cos_a - (-y1) * sin_a);
  q.y = p->y - MS_NINT(x1 * sin_a + (-y1) * cos_a);  

  if(poly) {
    line.point = (pointObj *)malloc(sizeof(pointObj)*5);
    line.numpoints = 5;

    x2 = x1 - buffer; /* ll */
    y2 = y1 + buffer;
    line.point[0].x = p->x + MS_NINT(x2 * cos_a - (-y2) * sin_a);
    line.point[0].y = p->y - MS_NINT(x2 * sin_a + (-y2) * cos_a);
    
    x2 = x1 - buffer; /* ul */
    y2 = y1 - h - buffer;
    line.point[1].x = p->x + MS_NINT(x2 * cos_a - (-y2) * sin_a);
    line.point[1].y = p->y - MS_NINT(x2 * sin_a + (-y2) * cos_a);
    
    x2 = x1 + w + buffer; /* ur */
    y2 = y1 - h - buffer;
    line.point[2].x = p->x + MS_NINT(x2 * cos_a - (-y2) * sin_a);
    line.point[2].y = p->y - MS_NINT(x2 * sin_a + (-y2) * cos_a);
    
    x2 = x1 + w + buffer; /* lr */
    y2 = y1 + buffer;
    line.point[3].x = p->x + MS_NINT(x2 * cos_a - (-y2) * sin_a);
    line.point[3].y = p->y - MS_NINT(x2 * sin_a + (-y2) * cos_a);
    
    line.point[4].x = line.point[0].x;
    line.point[4].y = line.point[0].y;
    
    msAddLine(poly, &line);
    free(line.point);
  }

  return(q);
}

/*
** Simply draws a label based on the label point and the supplied label object. 
*/
static int draw_text(gdImagePtr img, pointObj labelPnt, char *string, labelObj *label, fontSetObj *fontset)
{
  int x, y;

  if(!string)
    return(0); /* not an error, just don't want to do anything */

  if(strlen(string) == 0)
    return(0); /* not an error, just don't want to do anything */

  x = MS_NINT(labelPnt.x);
  y = MS_NINT(labelPnt.y);

  if(label->type == MS_TRUETYPE) {
    char *error=NULL, *font=NULL;
    int bbox[8];

#ifdef USE_TTF

    font = msLookupHashTable(fontset->fonts, label->font);
    if(!font) {
      msSetError(MS_TTFERR, NULL, "msGetLabelSize()");
      sprintf(ms_error.message, "Requested font (%s) not found.", label->font);
      return(-1);
    }

    if(label->outlinecolor >= 0) { /* handle the outline color */
      error = imageStringTTF(img, bbox, label->antialias*label->outlinecolor, font, label->size, label->angle, x-1, y-1, string, label->wrap);
      if(error) {
	msSetError(MS_TTFERR, error, "msDrawLabel()");
	return(-1);
      }
      imageStringTTF(img, bbox, label->antialias*label->outlinecolor, font, label->size, label->angle, x-1, y+1, string, label->wrap);
      imageStringTTF(img, bbox, label->antialias*label->outlinecolor, font, label->size, label->angle, x+1, y+1, string, label->wrap);
      imageStringTTF(img, bbox, label->antialias*label->outlinecolor, font, label->size, label->angle, x+1, y-1, string, label->wrap);
    }

    if(label->shadowcolor >= 0) { /* handle the shadow color */
      error = imageStringTTF(img, bbox, label->antialias*label->shadowcolor, font, label->size, label->angle, x+label->shadowsizex, y+label->shadowsizey, string, label->wrap);
      if(error) {
	msSetError(MS_TTFERR, error, "msDrawLabel()");
	return(-1);
      }
    }

    imageStringTTF(img, bbox, label->antialias*label->color, font, label->size, label->angle, x, y, string, label->wrap);
    
#else
    msSetError(MS_TTFERR, "TrueType font support is not available.", "msDrawLabel()");
    return(-1);
#endif

  } else { /* MS_BITMAP */
    char **token=NULL;
    int t, num_tokens;
    gdFontPtr fontPtr;
    
    if((fontPtr = msGetBitmapFont(label->size)) == NULL)
      return(-1);

    if((token = split(string, label->wrap, &(num_tokens))) == NULL)
      return(-1);
    
    y -= fontPtr->h*num_tokens;
    for(t=0; t<num_tokens; t++) { /* what's the longest token */
      if(label->outlinecolor >= 0) {
	gdImageString(img, fontPtr, x-1, y-1, token[t], label->outlinecolor);
	gdImageString(img, fontPtr, x-1, y+1, token[t], label->outlinecolor);
	gdImageString(img, fontPtr, x+1, y+1, token[t], label->outlinecolor);
	gdImageString(img, fontPtr, x+1, y-1, token[t], label->outlinecolor);
      }

      if(label->shadowcolor >= 0) { /* handle the shadow color */
	gdImageString(img, fontPtr, x+label->shadowsizex, y+label->shadowsizey, token[t], label->shadowcolor);
      }

      gdImageString(img, fontPtr, x, y, token[t], label->color);
      y += fontPtr->h; /* shift down */
    }
    
    msFreeCharArray(token, num_tokens);  
  }

  return(0);
}

int msDrawLabel(gdImagePtr img, mapObj *map, pointObj labelPnt, char *string, labelObj *label)
{
  if(!string)
    return(0); /* not an error, just don't want to do anything */

  if(strlen(string) == 0)
    return(0); /* not an error, just don't want to do anything */

  if(label->position >= 0) {
    pointObj p;
    int w, h;

    if(msGetLabelSize(string, label, &w, &h, &(map->fontset)) == -1)
      return(-1);
    p = get_metrics(&labelPnt, label->position, w, h, 0, 0, label->angle, 0, NULL);
    draw_text(img, p, string, label, &(map->fontset)); /* actually draw the label */
  } else {
    draw_text(img, labelPnt, string, label, &(map->fontset)); /* actually draw the label */
  }
  
  return(0);
}

/*
** is a label completely in the image (excluding label buffer)
*/ 
static int labelInImage(int width, int height, shapeObj *lpoly, int buffer)
{
  int i,j;

  for(i=0; i<lpoly->numlines; i++) {
    for(j=1; j<lpoly->line[i].numpoints; j++) {
      if(lpoly->line[i].point[j].x < -buffer) return(MS_FALSE);
      if(lpoly->line[i].point[j].x >= width+buffer) return(MS_FALSE);
      if(lpoly->line[i].point[j].y < -buffer) return(MS_FALSE);
      if(lpoly->line[i].point[j].y >= height+buffer) return(MS_FALSE);
    }
  }

  return(MS_TRUE);
}

static double dist(pointObj a, pointObj b)
{
  double d;

  d = sqrt((pow((a.x-b.x),2) + pow((a.y-b.y),2)));
  return(d);
}
   
int msDrawLabelCache(gdImagePtr img, mapObj *map)
{
  pointObj p;
  int i, j, l;

  int w, h;

  labelObj label;
  labelCacheMemberObj *cachePtr=NULL;
  classObj *classPtr=NULL;
  layerObj *layerPtr=NULL;

  int draw_marker, has_marker;
  int marker_width, marker_height;
  int marker_offset_x, marker_offset_y;
  rectObj marker_rect;

  int start_position;

  /* for(l=0; l<map->labelcache.numlabels; l++) { */
  for(l=map->labelcache.numlabels-1; l>=0; l--) {

    cachePtr = &(map->labelcache.labels[l]); /* point to right spot in cache */

    layerPtr = &(map->layers[cachePtr->layeridx]); /* set a couple of other pointers, avoids nasty references */
    classPtr = &(layerPtr->class[cachePtr->classidx]);

    if(!cachePtr->string)
      continue; /* not an error, just don't want to do anything */
    
    if(strlen(cachePtr->string) == 0)
      continue; /* not an error, just don't want to do anything */

    label = classPtr->label;
    label.size = cachePtr->size;
    label.angle = cachePtr->angle;

    if(msGetLabelSize(cachePtr->string, &label, &w, &h, &(map->fontset)) == -1)
      return(-1);

    if((cachePtr->featuresize > 0) && (label.autominfeaturesize == MS_TRUE) && (w > cachePtr->featuresize))
      continue; /* label too large relative to the feature */
    
    has_marker = draw_marker = marker_offset_x = marker_offset_y = 0; /* assume no marker */
    if((layerPtr->type == MS_ANNOTATION || layerPtr->type == MS_POINT) && (classPtr->color > 0 || classPtr->outlinecolor > 0)) { /* there *is* a marker */

      msGetMarkerSize(&(map->markerset), classPtr, &marker_width, &marker_height);
      marker_offset_x = MS_NINT(marker_width/2.0);
      marker_offset_y = MS_NINT(marker_height/2.0);

      marker_rect.minx = cachePtr->point.x - MS_NINT(.5 * marker_width);
      marker_rect.miny = cachePtr->point.y - MS_NINT(.5 * marker_height);
      marker_rect.maxx = cachePtr->point.x + MS_NINT(.5 * marker_width);
      marker_rect.maxy = cachePtr->point.y + MS_NINT(.5 * marker_height);

      if(layerPtr->type == MS_ANNOTATION) draw_marker = 1; /* actually draw the marker */
      if(layerPtr->type == MS_POINT) has_marker = 1; /* markers already drawn */
    }
    
    if(label.position == MS_AUTO) {
      
      if(layerPtr->type == MS_LINE)
	start_position = 6; /* positions 6 and 7 are UC and LC */
      else
	start_position = 0;
      
      for(j=start_position; j<=7; j++) { /* loop through the outer label positions */	  

	msFreeShape(cachePtr->poly);
	cachePtr->status = MS_TRUE; /* assume label *can* be drawn */

	p = get_metrics(&(cachePtr->point), j, w, h, (marker_offset_x + label.offsetx), (marker_offset_y + label.offsety), label.angle, label.buffer, cachePtr->poly);

	if(draw_marker) 
	  msRect2Polygon(marker_rect, cachePtr->poly); /* save marker bounding polygon, part of overlap tests */
	
	if(label.partials == MS_FALSE) { /* check against image first */
	  if(labelInImage(img->sx, img->sy, cachePtr->poly, label.buffer) == MS_FALSE) {
	    cachePtr->status = MS_FALSE;
	    continue; /* next position */
	  }
	}
	
	/* for(i=l+1; i<map->labelcache.numlabels; i++) { */
	for(i=0; i<map->labelcache.numlabels; i++) {
	  if(i!=l && map->labelcache.labels[i].status == MS_TRUE) { /* compare bounding polygons and check for duplicates */
	    
	    if((label.mindistance != -1) && (cachePtr->classidx == map->labelcache.labels[i].classidx) && (strcmp(cachePtr->string,map->labelcache.labels[i].string) == 0) && (dist(cachePtr->point, map->labelcache.labels[i].point) <= label.mindistance)) { /* label is a duplicate */
	      cachePtr->status = MS_FALSE;
	      break;
	    }
	    
	    if(msIntersectPolygons(map->labelcache.labels[i].poly, cachePtr->poly) == MS_TRUE) { /* polys intersect */
	      cachePtr->status = MS_FALSE;
	      break;
	    }
	  }
	}	

	if(cachePtr->status == MS_TRUE) /* found a placement */
	  break;
      } /* next position */      

    } else {

      cachePtr->status = MS_TRUE; /* assume label *can* be drawn */

      if(has_marker) /* marker already drawn, just need to consider the text */
	msFreeShape(cachePtr->poly);

      p = get_metrics(&(cachePtr->point), label.position, w, h, (marker_offset_x + label.offsetx), (marker_offset_y + label.offsety), label.angle, label.buffer, cachePtr->poly);

      if(draw_marker)
	msRect2Polygon(marker_rect, cachePtr->poly); /* save marker bounding polygon, part of overlap tests */
      
      if(label.partials == MS_FALSE) { /* check against image first */
	if(labelInImage(img->sx, img->sy, cachePtr->poly, label.buffer) == MS_FALSE) {
	  cachePtr->status = MS_FALSE;
	} else {
	  /* for(i=l+1; i<map->labelcache.numlabels; i++) { */
	  for(i=0; i<map->labelcache.numlabels; i++) {
	    if(i!=l && map->labelcache.labels[i].status == MS_TRUE) { /* compare bounding polygons and check for duplicates */		

	      if((label.mindistance != -1) && (cachePtr->classidx == map->labelcache.labels[i].classidx) && (strcmp(cachePtr->string,map->labelcache.labels[i].string) == 0) && (dist(cachePtr->point, map->labelcache.labels[i].point) <= label.mindistance)) { /* label is a duplicate */
		cachePtr->status = MS_FALSE;
		break;
	      }
	      
	      if(msIntersectPolygons(map->labelcache.labels[i].poly, cachePtr->poly) == MS_TRUE) { /* polys intersect */
		cachePtr->status = MS_FALSE;
		break;
	      }
	    }
	  }
	}
      } else {
	/* for(i=l+1; i<map->labelcache.numlabels; i++) { */
	for(i=0; i<map->labelcache.numlabels; i++) {
	  if(i!=l && map->labelcache.labels[i].status == MS_TRUE) { /* compare bounding polygons and check for duplcates */

	    if((label.mindistance != -1) && (cachePtr->classidx == map->labelcache.labels[i].classidx) && (strcmp(cachePtr->string,map->labelcache.labels[i].string) == 0) && (dist(cachePtr->point, map->labelcache.labels[i].point) <= label.mindistance)) { /* label is a duplicate */
	      cachePtr->status = MS_FALSE;
	      break;
	    }
	    
	    if(msIntersectPolygons(map->labelcache.labels[i].poly, cachePtr->poly) == MS_TRUE) { /* polys intersect */
	      cachePtr->status = MS_FALSE;
	      break;
	    }
	  }
	}
      }      
    } /* end position if-then-else */

    /* msImagePolygon(img, cachePtr->poly, 1); */

    if(!cachePtr->status) {
      if(has_marker) {
	msFreeShape(cachePtr->poly); /* delete the text poly */
	msRect2Polygon(marker_rect, cachePtr->poly); /* store the marker poly */
	cachePtr->status = MS_TRUE; /* still want to consider the marker */
      }
      continue; /* next label */
    }

    if(has_marker)
      msRect2Polygon(marker_rect, cachePtr->poly); /* save marker bounding polygon, part of future overlap tests */

    if(draw_marker) /* need to draw a marker */
      msDrawMarkerSymbol(&(map->markerset), img, &(cachePtr->point), classPtr);

    draw_text(img, p, cachePtr->string, &label, &(map->fontset)); /* actually draw the label */

  } /* next in cache */

  return(0);
}
