#include "map.h"
#include "mapparser.h"

/*
** This function fills the items and data buffers for a given join (record based).
** Now handles one-to-many joins correctly. Cleans up previous join before allocating
** space for the new one. 
*/ 
int msJoinDBFTables(joinObj *join) {
  int i, j, idx;
  DBFHandle hDBF;
  int nrecs, *ids=NULL;

  /* first open the lookup table file */
  if((hDBF = DBFOpen(join->table, "rb")) == NULL) {
    msSetError(MS_IOERR, NULL, "msJoinDBFTables()");
    return(-1);
  }

  if((idx = msGetItemIndex(hDBF, join->to)) == -1) {    
    msSetError(MS_DBFERR, NULL, "msJoinDBFTables()");
    sprintf(ms_error.message, "Item %s not found.", join->to);
    DBFClose(hDBF);
    return(-1);
  }

  /*
  ** Now we need to pull the record and item names in to the join buffers
  */
  join->numitems =  DBFGetFieldCount(hDBF);
  if(!join->items) {
    join->items = msGetDBFItems(hDBF);
    if(!join->items)
      return(-1);
  }

  nrecs = DBFGetRecordCount(hDBF);

  if(join->type == MS_SINGLE) { /* only one row */
    
    if((join->data = (char ***)malloc(sizeof(char **))) == NULL) {
      msSetError(MS_MEMERR, NULL, "msJoinDBFTables()");
      return(-1);
    }
    
    for(i=0; i<nrecs; i++) { /* find a match */
      if(strcmp(join->match, DBFReadStringAttribute(hDBF, i, idx)) == 0)
	break;
    }  
    
    if(i == nrecs) { /* just return zero length strings */
      if((join->data[0] = (char **)malloc(sizeof(char *)*join->numitems)) == NULL) {
	msSetError(MS_MEMERR, NULL, "msJoinDBFTables()");
	return(-1);
      }
      for(i=0; i<join->numitems; i++)
	join->data[0][i] = strdup("\0"); /* intialize to zero length strings */
    } else {
      if((join->data[0] = msGetDBFValues(hDBF,i)) == NULL)
	return(-1);
    }

  } else {

    if(join->data) { /* free old data */
      for(i=0; i<join->numrecords; i++)
	msFreeCharArray(join->data[i], join->numitems);
      free(join->data);
      join->numrecords = 0;
    }

    ids = (int *)malloc(sizeof(int)*nrecs);
    if(!ids) {
      msSetError(MS_MEMERR, NULL, "msJoinDBFTables()");
      return(-1);
    }
    
    j=0;
    for(i=0; i<nrecs; i++) { /* find the matches, save ids */
      if(strcmp(join->match, DBFReadStringAttribute(hDBF, i, idx)) == 0) {
	ids[j] = i;
	j++;
      }
    }
  
    join->numrecords = j;

    if(join->numrecords > 0) { /* save em */
      if((join->data = (char ***)malloc(sizeof(char **)*join->numrecords)) == NULL) {
	msSetError(MS_MEMERR, NULL, "msJoinDBFTables()");
	free(ids);
	return(-1);
      }

      for(i=0; i<join->numrecords; i++) {
	join->data[i] = (char **)malloc(sizeof(char *)*join->numitems);
	if(!join->data[i]) {
	  msSetError(MS_MEMERR, NULL, "msJoinDBFTables()");
	  free(ids);
	  return(-1);
	}

	join->data[i] = msGetDBFValues(hDBF,ids[i]);
	if(!join->data[i]) {
	  free(ids);
	  return(-1);
	}
      }      
    }

    free(ids);
  }

  DBFClose(hDBF);
  return(0);
}

/*
** Allocates memory for query result set
*/
static queryResultObj *newResults(int n)
{
  queryResultObj *results;
  int i;

  if((results = (queryResultObj *)malloc(sizeof(queryResultObj)*n)) == NULL) {
    msSetError(MS_MEMERR, NULL, "newResults()"); 
    return(NULL);
  }

  for(i=0; i<n; i++) { /* initialize a few things */
    results[i].status = NULL;
    results[i].index = NULL;
    results[i].numresults = -1;
  }

  return(results);
}

/*
** Frees memory for a query result set
*/
void msFreeQueryResults(queryResultObj *results, int n)
{
  int i;

  for(i=0; i<n; i++) {
    free(results[i].status);
    free(results[i].index);
  }
  free(results);
}

extern int yyresult; // result of parsing, true/false
extern int yystate;
extern char *yystring;
extern int yyparse();

static int shpGetQuery(DBFHandle hDBF, layerObj *layer, int record, int item)
{
  int i,j;
  int numitems=0;
  char *tmpstr=NULL, **values=NULL;
  int found=MS_FALSE;

  if((layer->numqueries == 1) && !(layer->query[0].expression.string)) /* no need to do lookup */
    return(0);

  for(i=0; i<layer->numqueries; i++) {
    switch(layer->query[i].expression.type) {
    case(MS_STRING):
      if(strcmp(layer->query[i].expression.string, DBFReadStringAttribute(hDBF, record, item)) == 0) /* got a match */
	found=MS_TRUE;
      break;
    case(MS_EXPRESSION):

      tmpstr = strdup(layer->query[i].expression.string);

      if(!values) {
	numitems = DBFGetFieldCount(hDBF);
	values = msGetDBFValues(hDBF, record);
      }

      if(!(layer->query[i].expression.items)) { // build cache of item replacements	
	char **items=NULL, substr[19];

	numitems = DBFGetFieldCount(hDBF);
	items = msGetDBFItems(hDBF);
	
	layer->query[i].expression.items = (char **)malloc(numitems);
	layer->query[i].expression.indexes = (int *)malloc(numitems);

	for(j=0; j<numitems; j++) {
	  sprintf(substr, "[%s]", items[j]);
	  if(strstr(tmpstr, substr) != NULL) {
	    layer->query[i].expression.indexes[layer->query[i].expression.numitems] = j;
	    layer->query[i].expression.items[layer->query[i].expression.numitems] = strdup(substr);
	    layer->query[i].expression.numitems++;
	  }
	}

	msFreeCharArray(items,numitems);
      }

      for(j=0; j<layer->query[i].expression.numitems; j++)
	tmpstr = gsub(tmpstr, layer->query[i].expression.items[j], values[(int)layer->query[i].expression.indexes[j]]);

      yystate = 4; yystring = tmpstr;
      if(yyparse() != 0)
	return(-1);

      free(tmpstr);

      if(yyresult) /* got a match */
	found=MS_TRUE;
      break;
    case(MS_REGEX):
      if(regexec(&(layer->query[i].expression.regex), DBFReadStringAttribute(hDBF, record, item), 0, NULL, 0) == 0) /* got a match */
	found=MS_TRUE;
      break;
    }

    if(found) {
      if(values) msFreeCharArray(values,numitems);
      return(i);
    }
  }

  return(-1); /* not found */
}

/*
** This function builds a result set for a set of layers based on an item/value
** pairing. The value is actually a regular expression and all matching results
** are returned (depending on mode).
*/ 
queryResultObj *msQueryUsingItem(mapObj *map, char *layer, int mode, char *item, char *value)
{
  int i, l; /* counters */
  int start, stop=0;
  regex_t re; /* expression to be matched */
  queryResultObj *results=NULL;
  int itemIndex;
  shapefileObj shpfile;
  int index, queryItemIndex;
  char *status=NULL;

  if(regcomp(&re, value, REG_EXTENDED|REG_NOSUB) != 0) {
   msSetError(MS_REGEXERR, NULL, "msFindRecords()");
   return(NULL);
  }  

  /*
  ** Do we have query layer, if not we need to search all layers
  */
  if(layer != NULL) {
    if((start = msGetLayerIndex(map, layer)) == -1) {
      msSetError(MS_MISCERR, NULL, "msQueryUsingItem()");
      sprintf(ms_error.message, "Undefined layer %s.", layer);
      return(NULL);
    }
    stop = start;
  } else {
    start = map->numlayers-1;
  }

  results = newResults(map->numlayers); /* results exist for each layer */

  for(l=start; l>=stop; l--) {
     
    if((map->layers[l].status == MS_OFF) || (map->layers[l].numqueries == 0) || (map->layers[l].tileindex != NULL))
      continue;

    results[l].numresults = 0; /* init, -1 means layer not searched */    

    if(msOpenSHPFile(&shpfile, map->shapepath, map->tile, map->layers[l].data) == -1) return(NULL);

    /*
    ** allocate space for status and query index arrays
    */
    results[l].status = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].status) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }
    
    results[l].index = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].index) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }

    queryItemIndex = msGetItemIndex(shpfile.hDBF, map->layers[l].queryitem);
    
    if((itemIndex = msGetItemIndex(shpfile.hDBF, item)) == -1) { /* requested item doesn't exist in this table */
      msCloseSHPFile(&shpfile);
      continue;
    }

    if(map->extent.minx != map->extent.maxx) { /* use extent as first cut */
#ifdef USE_PROJ
      status = msWhichShapesProj(&shpfile, map->extent, &(map->layers[l].projection), &(map->projection));
#else	    
      status = msWhichShapes(&shpfile, map->extent);
#endif      
    } else {
      status = (char *)malloc(shpfile.numshapes);
      for(i=0; i<shpfile.numshapes; i++) status[i] = 1;
    }

    if(!status) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }
    
    for(i=0; i<shpfile.numshapes; i++) {
      if(status[i] == 0)
	continue; /* next shape */

      if(regexec(&re, DBFReadStringAttribute(shpfile.hDBF, i, itemIndex), 0, NULL, 0) == 0) { /* match */
	
	index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), i, queryItemIndex);
	if((index < 0) || (index >= map->layers[l].numqueries))
	  continue;
	
	results[l].status[i] = 1;
	results[l].index[i] = index;
	results[l].numresults++;

	if(mode == MS_SINGLE) { /* no need to go any further */
	  regfree(&re);
	  free(status);
	  msCloseSHPFile(&shpfile);
	  return(results);
	}
      }
    }

    free(status);
    msCloseSHPFile(&shpfile);
  } /* next layer */

  regfree(&re); /* clear expression */

  for(l=start; l>=stop; l--) {
    if(results[l].numresults > 0) /* there are at least some results */
      return(results);
  }
  
  msSetError(MS_NOTFOUND, "No matching record(s) found.", "msQueryUsingItem()"); 
  msFreeQueryResults(results, map->numlayers);
  return(NULL); 
}

/*
** This function builds a result set based on a search rectangle. The
** rectangle is intersected with all *active* layers shape by shape.
*/
queryResultObj *msQueryUsingRect(mapObj *map, char *layer, rectObj rect) {
  int i,j,l; /* counters */
  int start, stop=0;
  queryResultObj *results=NULL;
  char *status=NULL;
  shapefileObj shpfile;
  int index, queryItemIndex;
  rectObj shprect;

  shapeObj shape={0,NULL}, search_shape={0,NULL};

  if(layer) {
    if((start = msGetLayerIndex(map, layer)) == -1) {
      msSetError(MS_MISCERR, NULL, "msQueryUsingRect()");
      sprintf(ms_error.message, "Undefined layer %s.", layer);
      return(NULL);
    }
    stop = start;
  } else {
    start = map->numlayers-1;
  }

  msRect2Polygon(rect, &search_shape);

  results = newResults(map->numlayers); /* results exist for each layer */

  for(l=start; l>=stop; l--) {

    if((map->layers[l].status == MS_OFF) || (map->layers[l].numqueries == 0) || (map->layers[l].tileindex != NULL))
      continue;

    if(map->scale > 0) {
      if((map->layers[l].maxscale > 0) && (map->scale > map->layers[l].maxscale)) continue;
      if((map->layers[l].minscale > 0) && (map->scale <= map->layers[l].minscale)) continue;
    }

    results[l].numresults = 0; /* init, -1 means layer not searched */

    if(msOpenSHPFile(&shpfile,  map->shapepath, map->tile, map->layers[l].data) == -1) return(NULL);
       
#ifdef USE_PROJ
    status = msWhichShapesProj(&shpfile, rect, &(map->layers[l].projection), &(map->projection));
#else	    
    status = msWhichShapes(&shpfile, rect);    
#endif
    if(!status) {
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }

    /*
    ** allocate space for status and query index arrays
    */
    results[l].status = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].status) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }
    
    results[l].index = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].index) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }

    queryItemIndex = msGetItemIndex(shpfile.hDBF, map->layers[l].queryitem);

    for(i=0;i<shpfile.numshapes;i++) { /* step through the shapes */ 

      if(!status[i]) continue; /* next shape */

#ifdef USE_PROJ
      msReadBoundsProj(shpfile.hSHP, i, &shprect, &(map->layers[l].projection), &(map->projection));
#else
      msReadBounds(shpfile.hSHP, i, &shprect);
#endif

      if(msRectContained(shprect, rect) == MS_TRUE) { /* if the whole shape is in, don't intersect */
	index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), i, queryItemIndex);
	if((index >= 0) && (index < map->layers[l].numqueries)) {
	  results[l].status[i] = 1;
	  results[l].index[i] = index;
	  results[l].numresults++;
	}
	continue; /* next shape */
      }

      switch(shpfile.type) { /* make sure shape actually intersects the rect */
      case SHPT_POINT:
      case SHPT_MULTIPOINT:

#ifdef USE_PROJ
	msReadShapeProj(shpfile.hSHP, i, &shape, &(map->layers[l].projection), &(map->projection));
#else	    
	msReadShape(shpfile.hSHP, i, &shape);
#endif

	for(j=0;j<shape.line[0].numpoints;j++) {
	  if(msPointInRect(shape.line[0].point[j], rect) == MS_TRUE) {
	    index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), i, queryItemIndex);
	    if((index >= 0) && (index < map->layers[l].numqueries)) {
	      results[l].status[i] = 1;
	      results[l].index[i] = index;
	      results[l].numresults++;
	    }
	    break;
	  }
	} /* next point */
	
	msFreeShape(&shape);
	
	break;
      case SHPT_ARC:

#ifdef USE_PROJ
	msReadShapeProj(shpfile.hSHP, i, &shape, &(map->layers[l].projection), &(map->projection));	    
#else
	msReadShape(shpfile.hSHP, i, &shape);	    
#endif
	
	if(msIntersectPolylinePolygon(&shape, &search_shape) == MS_TRUE) {
	  index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), i, queryItemIndex);
	  if((index >= 0) && (index < map->layers[l].numqueries)) {
	    results[l].status[i] = 1;
	    results[l].index[i] = index;
	    results[l].numresults++;
	  }	  
	}

	msFreeShape(&shape);
	break;
      case SHPT_POLYGON:

#ifdef USE_PROJ
	msReadShapeProj(shpfile.hSHP, i, &shape, &(map->layers[l].projection), &(map->projection));	    
#else
	msReadShape(shpfile.hSHP, i, &shape);	    
#endif
	  
	if(msIntersectPolygons(&shape, &search_shape) == MS_TRUE) {
	  index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), i, queryItemIndex);
	  if((index >= 0) && (index < map->layers[l].numqueries)) {
	    results[l].status[i] = 1;
	    results[l].index[i] = index;
	    results[l].numresults++;
	  }
	}

	msFreeShape(&shape);
	break;
      default:
	break;
      }
    } /* next shape */

    free(status);
    msCloseSHPFile(&shpfile);

  } /* next layer */
 
  msFreeShape(&search_shape);

  for(l=start; l>=stop; l--) {
    if(results[l].numresults > 0) /* there are at least some results */
      return(results);
  }
  
  msSetError(MS_NOTFOUND, "No matching record(s) found.", "msQueryUsingRect()"); 
  msFreeQueryResults(results, map->numlayers);
  return(NULL); 
}

/*
** This function searches multiple themes if necessary based on the state of a mapfile.
** If a query layer is supplied then only that layer is searched. Depending on mode it
** will return the first/closest feature or all features within a given tolerance. Alternatively
** a global buffer size can be specified in map units when an image based interface is not
** used as the front-end.
*/
queryResultObj *msQueryUsingPoint(mapObj *map, char *layer, int mode, pointObj p, double buffer)
{
  int i, l; /* counters */
  int start, stop=0;
  shapefileObj shpfile;
  double tolerance=0, d;
  rectObj rect;
  queryResultObj *results=NULL;
  char *status=NULL; 
  int index, queryItemIndex;

  shapeObj shape={0,NULL};

  /*
  ** Do we have query layer, if not we need to search all layers
  */
  if(layer) {
    if((start = msGetLayerIndex(map, layer)) == -1) {
      msSetError(MS_MISCERR, NULL, "msQueryUsingPoint()");
      sprintf(ms_error.message, "Undefined layer %s.", layer);
      return(NULL);
    }
    stop = start;
  } else {
    start = map->numlayers-1;
  }
  
  results = newResults(map->numlayers); /* results exist for each layer */
  
  for(l=start; l>=stop; l--) {
    
    if((map->layers[l].status == MS_OFF) || (map->layers[l].numqueries == 0) || (map->layers[l].tileindex != NULL))
      continue;

    if(map->scale > 0) {
      if((map->layers[l].maxscale > 0) && (map->scale > map->layers[l].maxscale)) continue;
      if((map->layers[l].minscale > 0) && (map->scale <= map->layers[l].minscale)) continue;
    }

    results[l].numresults = 0; /* init, -1 means layer not searched */

    if(msOpenSHPFile(&shpfile,  map->shapepath, map->tile, map->layers[l].data) == -1) return(NULL);
  
    if(buffer <= 0) /* use layer tolerance */ 
      tolerance = map->layers[l].tolerance * msAdjustExtent(&(map->extent), map->width, map->height);
    else /* use buffer distance */
      tolerance = buffer;

    rect.minx = p.x - tolerance;
    rect.maxx = p.x + tolerance;
    rect.miny = p.y - tolerance;
    rect.maxy = p.y + tolerance;

#ifdef USE_PROJ
    status = msWhichShapesProj(&shpfile, rect, &(map->layers[l].projection), &(map->projection));
#else	    
    status = msWhichShapes(&shpfile, rect);    
#endif
    if(!status) {
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }

    /*
    ** allocate space for status and query index arrays
    */
    results[l].status = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].status) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }
    
    results[l].index = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].index) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }

    queryItemIndex = msGetItemIndex(shpfile.hDBF, map->layers[l].queryitem); /* which item */
 
    for(i=0;i<shpfile.numshapes;i++) { /* for each shape */

      if(!status[i]) continue;
      
      if((shpfile.type == SHPT_POINT) || (shpfile.type == SHPT_MULTIPOINT)) {

#ifdef USE_PROJ
	msReadShapeProj(shpfile.hSHP, i, &shape, &(map->layers[l].projection), &(map->projection));	    
#else	    
	msReadShape(shpfile.hSHP, i, &shape);	    
#endif

	d = msDistPoints(p, &shape.line[0]);
      } else {

#ifdef USE_PROJ
	msReadShapeProj(shpfile.hSHP, i, &shape, &(map->layers[l].projection), &(map->projection));	    
#else
	msReadShape(shpfile.hSHP, i, &shape);	    
#endif

	if(shpfile.type == SHPT_ARC)
	  d = msDistPolyline(p, &shape);
	else /* polygon */
	  d = msDistPolygon(p, &shape);	  
      }
      
      if( d <= tolerance ) { /* found one */
	index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), i, queryItemIndex);
	if((index < 0) || (index >= map->layers[l].numqueries))
	  continue;
	
	if(mode == MS_SINGLE) {
	  results[l].numresults = 1;
	  tolerance = d;
	} else {
	  results[l].numresults++;
	}
	
	results[l].status[i] = 1;
	results[l].index[i] = index;
      }
      
      msFreeShape(&shape);
    } /* next shape */      
    
    free(status);
    msCloseSHPFile(&shpfile);
    
    if((results[l].numresults > 0) && (mode == MS_SINGLE)) /* no need to search further */
      return(results);
  } /* next layer */

  for(l=start; l>=stop; l--) {
    if(results[l].numresults > 0) /* there are at least some results */
      return(results);
  }

  msSetError(MS_NOTFOUND, "No matching record(s) found.", "msQueryUsingPoint()");
  msFreeQueryResults(results, map->numlayers);
  return(NULL);
}

/*
** Generates a result set based on a previous set of results. At present the results MUST
** be based on a polygon layer. Holes and disjoint polygons should now be handled correctly.. 
*/
queryResultObj *msQueryUsingFeatures(mapObj *map, char *layer, queryResultObj *search_results)
{ 
  int start, stop=0;
  queryResultObj *results=NULL;
  int l, i, j, search_layer=-1;
  shapefileObj search_shpfile, shpfile;
  char *status=NULL;
  int index, queryItemIndex;
  rectObj rect;

  shapeObj search_shape={0,NULL}, shape={0,NULL};

  /*
  ** Do we have query layer, if not we need to search all layers
  */
  if(layer) {
    if((start = msGetLayerIndex(map, layer)) == -1) {
      msSetError(MS_MISCERR, NULL, "msFindUsingFeatures()");
      sprintf(ms_error.message, "Undefined layer %s.", layer);
      return(NULL);
    }
    stop = start;
  } else {
    start = map->numlayers-1;
  }

  results = newResults(map->numlayers); /* results exist for each layer */

  /* 
  ** open shapefile to pull features from 
  */
  for(l=0;l<map->numlayers;l++) {
    if(search_results[l].numresults > 0) {

      if(msOpenSHPFile(&search_shpfile,  map->shapepath, map->tile, map->layers[l].data) == -1) return(NULL);

      search_layer = l; /* save the index value */

      /* 
      ** copy the search layer results into the new result set (NEED CHECKS ON MALLOC)
      */
      results[l].numresults = search_results[l].numresults;
      results[l].status = (char *)malloc(search_shpfile.numshapes);
      memmove(results[l].status, search_results[l].status, search_shpfile.numshapes);
      results[l].index = (char *)malloc(search_shpfile.numshapes);
      memmove(results[l].index, search_results[l].index, search_shpfile.numshapes);

      break;
    }
  }  

  for(l=start; l>=stop; l--) { /* each layer */
        
    if((l == search_layer) || (map->layers[l].status == MS_OFF) || (map->layers[l].numqueries == 0) || (map->layers[l].tileindex != NULL))
      continue;

    if(map->scale > 0) {
      if((map->layers[l].maxscale > 0) && (map->scale > map->layers[l].maxscale)) continue;
      if((map->layers[l].minscale > 0) && (map->scale <= map->layers[l].minscale)) continue;
    }

    results[l].numresults = 0; /* init, -1 means layer not searched */

    if(msOpenSHPFile(&shpfile,  map->shapepath, map->tile, map->layers[l].data) == -1) return(NULL);

    /*
    ** allocate space for status and query index arrays
    */
    results[l].status = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].status) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }
    
    results[l].index = (char *)calloc(shpfile.numshapes, sizeof(char));    
    if(!results[l].index) {
      msSetError(MS_MEMERR, NULL, "msQueryUsingItem()");
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }

    queryItemIndex = msGetItemIndex(shpfile.hDBF, map->layers[l].queryitem);

    for(i=0; i<search_shpfile.numshapes; i++) { /* each result shape */

      if(search_results[search_layer].status[i] == 0)
	continue;

      /*
      ** Retrieve shapes that *may* intersect the one we're working on
      */
#ifdef USE_PROJ
      msReadBoundsProj(search_shpfile.hSHP, i, &rect, &(map->layers[l].projection), &(map->projection));
#else
      msReadBounds(search_shpfile.hSHP, i, &rect);
#endif

#ifdef USE_PROJ
    status = msWhichShapesProj(&shpfile, rect, &(map->layers[l].projection), &(map->projection));
#else	    
    status = msWhichShapes(&shpfile, rect);    
#endif
    if(!status) {
      msCloseSHPFile(&shpfile);
      msFreeQueryResults(results, map->numlayers);
      return(NULL);
    }   

#ifdef USE_PROJ
    msReadShapeProj(search_shpfile.hSHP, i, &search_shape, &(map->layers[l].projection), &(map->projection));
#else
    msReadShape(search_shpfile.hSHP, i, &search_shape); /* read the search polygon */
#endif
    
    switch(search_shpfile.type) { /* may eventually support other types than polygon */
    case SHPT_POLYGON:
      
      switch(shpfile.type) {
      case SHPT_POINT:
      case SHPT_MULTIPOINT:
	
	for(j=0; j<shpfile.numshapes; j++) {
	  
	  if(!status[j]) continue;
	  
	  if(results[l].status[j]) continue; /* already found this feature */
	  
#ifdef USE_PROJ
	  msReadShapeProj(shpfile.hSHP, j, &shape, &(map->layers[l].projection), &(map->projection));
#else	    
	  msReadShape(shpfile.hSHP, j, &shape);	    
#endif
	  
	  if(msIntersectPointsPolygon(&shape.line[0], &search_shape) == MS_TRUE) {
	    index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), j, queryItemIndex);
	    if((index >= 0) && (index < map->layers[l].numqueries)) {
	      results[l].status[j] = 1;
	      results[l].index[j] = index;
	      results[l].numresults++;
	    }
	  }	    
	  
	  msFreeShape(&shape);	
	}	  
	
	break; /* next search shape */
      case SHPT_ARC:
	
	for(j=0; j<shpfile.numshapes; j++) {
	  
	  if(!status[j]) continue;

	  if(results[l].status[j]) continue; /* already found this feature */
	  
#ifdef USE_PROJ
	  msReadShapeProj(shpfile.hSHP, j, &shape, &(map->layers[l].projection), &(map->projection));	    
#else
	  msReadShape(shpfile.hSHP, j, &shape);	    
#endif
	  
	  if(msIntersectPolylinePolygon(&shape, &search_shape) == MS_TRUE) {
	    index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), j, queryItemIndex);
	    if((index >= 0) && (index < map->layers[l].numqueries)) {
	      results[l].status[j] = 1;
	      results[l].index[j] = index;
	      results[l].numresults++;
	    }		
	  }
	  
	  msFreeShape(&shape);
	}
	
	break; /* next shape */
      case SHPT_POLYGON:
	
	for(j=0; j<shpfile.numshapes; j++) {
	  
	  if(!status[j]) continue;
	  
	  if(results[l].status[j]) continue; /* already found this feature */

#ifdef USE_PROJ
	  msReadShapeProj(shpfile.hSHP, j, &shape, &(map->layers[l].projection), &(map->projection));	    
#else
	  msReadShape(shpfile.hSHP, j, &shape);	    
#endif
	  
	  if(msIntersectPolygons(&shape, &search_shape) == MS_TRUE) {
	    index = shpGetQuery(shpfile.hDBF, &(map->layers[l]), j, queryItemIndex);
	    if((index >= 0) && (index < map->layers[l].numqueries)) {
	      results[l].status[j] = 1;
	      results[l].index[j] = index;
	      results[l].numresults++;
	    }
	  }
	  msFreeShape(&shape);
	}
	break; /* next result shape/part */
      } /* end shpfile switch */
      break;
    default:
      msSetError(MS_MISCERR, "Selection layer must contain polygons.", "msQueryUsingFeatures()");
      msCloseSHPFile(&search_shpfile);
      return(NULL);
    } /* end search_shpfile switch */
    
    free(status);
    msFreeShape(&search_shape);
    
    } /* next result shape */
    
    msCloseSHPFile(&shpfile);
    
  } /* next layer */

  msCloseSHPFile(&search_shpfile); /* done with selection layer */
  
  for(l=start; l>=stop; l--) {
    if(results[l].numresults > 0) /* there are at least some results */
      return(results);
  }
  
  msSetError(MS_NOTFOUND, "No matching record(s) found.", "msQueryUsingFeatures()");
  msFreeQueryResults(results, map->numlayers);
  return(NULL); 
}
