//
// mapscript.i: SWIG interface file for MapServer scripting extension called MapScript.
//

%module mapscript
%{
#include "../map.h"
extern struct featureObj *addFeature(struct featureObj *list);
extern struct featureObj *initFeature();
%}

%include typemaps.i
%include constraints.i

// grab mapserver declarations
%include "../map.h"
%include "../maperror.h"
%include "../mapprimitive.h"
%include "../mapsymbol.h"
%include "../mapshape.h"

%apply Pointer NONNULL { gdImagePtr img };

//
// class extensions for mapObj
//
%addmethods mapObj {
  mapObj(char *filename) {
    return msLoadMap(filename);
  }

  ~mapObj() {
    msFreeMap(self);
  }

  layerObj *getLayer(int i) {
    if(i >= 0 && i < self->numlayers)	
      return &(self->layers[i]); /* returns an EXISTING layer */
    else
      return NULL;
  }

  layerObj *getLayerByName(char *name) {
    int i;

    i = msGetLayerIndex(self, name);

    if(i != -1)
      return &(self->layers[i]); /* returns an EXISTING layer */
    else
      return NULL;
  }

  int addColor(int r, int g, int b) {
    return msAddColor(self, r, g, b);
  }

  int getSymbolByName(int type, char *name) {
    switch (type) {
    case(MS_MARKERSET):
      return msGetSymbolIndex(&self->markerset, name);
    case(MS_LINESET):
      return msGetSymbolIndex(&self->lineset, name);
    case(MS_SHADESET):
      return msGetSymbolIndex(&self->shadeset, name);
    default:
      return -1;
    }
  }

  gdImagePtr prepareImage() {
    gdImagePtr img;


    img = gdImageCreate(self->width, self->height);
    if(!img)
      return NULL;
  
    if(msLoadPalette(img, &(self->palette), self->imagecolor) == -1)
      return NULL;
  
    self->cellsize = msAdjustExtent(&(self->extent), self->width, self->height);
    self->scale = msCalculateScale(self->extent, self->units, self->width, self->height);

    return img;
  }

  gdImagePtr draw() {
    return msDrawMap(self);
  }

  gdImagePtr drawLegend() {
    return msDrawLegend(self);
  }

  gdImagePtr drawScalebar() {
    return msDrawScalebar(self);
  }

  gdImagePtr drawReferenceMap() {
    return msDrawReferenceMap(self);
  }

  int drawLabelCache(gdImagePtr img) {
    return msDrawLabelCache(img, self);
  }

  labelCacheMemberObj *nextLabel() {
    static int i=0;
    if(i<self->labelcache.numlabels)
      return &(self->labelcache.labels[i++]);
    else
      return NULL;	
  }
}


//
// class extensions for layerObj, always within the context of a map
//
%addmethods layerObj {
  layerObj(mapObj *map) {
    if(map->numlayers == MS_MAXLAYERS) /* no room */
      return(NULL);

    if(initLayer(&(map->layers[map->numlayers])) == -1)
      return(NULL);

    map->layers[map->numlayers].index = map->numlayers;
    map->numlayers++;

    return &(map->layers[map->numlayers-1]);
  }

  ~layerObj() {
    return; /* do nothing, map deconstrutor takes care of it all */
  }

  classObj *getClass(int i) { /* returns an EXISTING class */
    if(i >= 0 && i < self->numclasses)
      return &(self->class[i]); 
    else
      return(NULL);
  }

  int draw(mapObj *map, gdImagePtr img) {
    if(self->features) {
      return msDrawFeatures(map, self, img);
    } else {
      if(self->type == MS_RASTER) {
        return msDrawRaster(map, self, img);
      } else {
	return msDrawShapefile(map, self, img, NULL);
      }
    }
  }
}

//
// class extensions for classObj, always within the context of a layer
//
%addmethods classObj {
  classObj(layerObj *layer) {
    if(layer->numclasses == MS_MAXCLASSES) /* no room */
      return NULL;

    if(initClass(&(layer->class[layer->numclasses])) == -1)
      return NULL;

    layer->numclasses++;

    return &(layer->class[layer->numclasses-1]);
  }

  ~classObj() {
    return; /* do nothing, map deconstrutor takes care of it all */
  }
}

//
// class extensions for featureObj, always within the context of a layer
//
%addmethods featureObj {
  featureObj(layerObj *layer) {
    if(!layer->features) { /* new feature list */
      layer->features = initFeature();
      return layer->features;
    } else {
      return addFeature(layer->features);
    }
  }

  ~featureObj() {
    return; /* do nothing, map deconstrutor takes care of it all */
  }

  int add(lineObj *p) {
    return msAddLine(&self->shape, p);
  }
}

//
// class extensions for pointObj, useful many places
//
%addmethods pointObj {
  pointObj() {
    return (pointObj *)malloc(sizeof(pointObj));
  }

  ~pointObj() {
    free(self);
  }

  int draw(mapObj *map, layerObj *layer, gdImagePtr img, char *class_string, char *label_string) {
    return msDrawPoint(map, layer, self, img, class_string, label_string);
  }
}

//
// class extensions for lineObj (eg. a line or group of points), useful many places
//
%addmethods lineObj {
  lineObj() {
    lineObj *line;

    line = (lineObj *)malloc(sizeof(lineObj));
    if(!line)
      return(NULL);

    line->numpoints=0;
    line->point=NULL;

    return line;
  }

  ~lineObj() {
    free(self->point);
    free(self);		
  }

  pointObj *get(int i) {
    if(i<0 || i>=self->numpoints)
      return NULL;
    else
      return &(self->point[i]);
  }

  int add(pointObj *p) {
    if(self->numpoints == 0) { /* new */	
      self->point = (pointObj *)malloc(sizeof(pointObj));      
      if(!self->point)
	return -1;
    } else { /* extend array */
      self->point = (pointObj *)realloc(self->point, sizeof(pointObj)*(self->numpoints+1));
      if(!self->point)
	return -1;
    }

    self->point[self->numpoints].x = p->x;
    self->point[self->numpoints].y = p->y;
    self->numpoints++;

    return 0;
  }
}

//
// class extensions for shapeObj
//
%addmethods shapeObj {
  shapeObj() {
    shapeObj *shape;

    shape = (shapeObj *)malloc(sizeof(shapeObj));
    if(!shape)
      return(NULL);

    shape->numlines=0;
    shape->line=NULL;

    return shape;
  }

  ~shapeObj() {
    msFreeShape(self);
    free(self);		
  }

  lineObj *get(int i) {
    if(i<0 || i>=self->numlines)
      return NULL;
    else
      return &(self->line[i]);
  }

  int add(lineObj *line) {
    return msAddLine(self, line);
  }

  int draw(mapObj *map, layerObj *layer, gdImagePtr img, char *class_string, char *label_string) {
    return msDrawShape(map, layer, self, img, class_string, label_string);
  }
}

//
// class extensions for rectObj
//
%addmethods rectObj {
  rectObj() {	
    rectObj *rect;

    rect = (rectObj *)calloc(1, sizeof(rectObj));
    if(!rect)
      return(NULL);
    
    return(rect);    	
  }

  ~rectObj() {
    free(self);
  }

  int draw(mapObj *map, layerObj *layer, gdImagePtr img, char *class_string, char *label_string) {
    shapeObj shape;

    msRect2Polygon(*self, &shape);
    msDrawShape(map, layer, &shape, img, class_string, label_string);
    msFreeShape(&shape);
    
    return 0;
  }
}

//
// class extensions for shapefileObj
//
%addmethods shapefileObj {
  shapefileObj(char *filename, int type) {    
    shapefileObj *shapefile;
    int status;

    shapefile = (shapefileObj *)malloc(sizeof(shapefileObj));
    if(!shapefile)
      return(NULL);

    if(type == -1)
      status = msOpenSHPFile(shapefile, NULL, NULL, filename);
    else
      status = msCreateSHPFile(shapefile, filename, type);
    if(status == -1) {
      msCloseSHPFile(shapefile);
      free(shapefile);
      return(NULL);
    }
 
    return(shapefile);
  }

  ~shapefileObj() {
    msCloseSHPFile(self);
    free(self);  
  }	

  int get(int i, shapeObj *shape) {
    if(i<0 || i>=self->numshapes)
      return -1;

    msFreeShape(shape); /* frees all lines and points before re-filling */
    msReadShape(self->hSHP, i, shape);

    return 0;
  }

  void getExtent(int i, rectObj *rect) {
    msReadBounds( self->hSHP, i, rect );
  }

  int add(shapeObj *shape) {
    return msWriteShape(self->hSHP, shape);	
  }	
}

